#from youtube_autonomous.segments.enums import SegmentField


class ElementData:
    """
    A class to handle any project Element information.
    """
    # Narration
    audio_narration_filename: str
    narration_text: str
    voice: str
    # Text
    text: str
    # Specific duration
    duration: float
    # Source (filename or url)
    filename: str
    url: str
    # Source or identifier (keywords)
    keywords: str

    # During building
    calculated_duration: float
    start: float
    mode: str
    audio_filename: str
    video_filename: str
    full_filename: str
    transcription = None
    shortcodes = []

    def __init__(self, json: dict):
        # TODO: Apply 'SegmentField' instead of this strings, please
        self.audio_narration_filename = json.get('audio_narration_filename', '')
        self.narration_text = json.get('narration_text', '')
        self.voice = json.get('voice', '')

        self.text = json.get('text', '')

        self.duration = json.get('duration', None)
        
        self.filename = json.get('filename', '')
        self.url = json.get('url', '')

        self.keywords = json.get('keywords', '')

        self.calculated_duration = json.get('calculated_duration', None)
        self.start = json.get('start', None)
        self.mode = json.get('mode', '')
        self.audio_filename = json.get('audio_filename', '')
        self.video_filename = json.get('video_filename', '')
        self.full_filename = json.get('full_filename', '')

    def as_enum(self):
        """
        Returns the information
        """
        # TODO: (?)


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enums import SegmentField
from youtube_autonomous.segments.builder.config import MAX_DURATION_PER_IMAGE, MIN_DURATION_PER_IMAGE
from youtube_autonomous.segments.builder.ai import create_ai_image
from moviepy.editor import concatenate_videoclips
from typing import Union


class AIImageElementBuilder(ElementBuilder):
    """
    Builder to obtain videos generated with AI image (or images) 
    turned into video.
    """
    @classmethod
    def get_images_array(cls, keywords: str, duration: Union[float, int]):
        """
        Builds an array containing as much image descriptions as
        needed to build a video of 'duration' seconds according
        to our MAX_DURATION_PER_IMAGE configuration settings.

        Each image of the array will have 'keywords' and 'duration'
        fields to be able to build the video.

        TODO: The 'keywords' field is not necessary in each array
        element, but...
        """
        ElementParameterValidator.validate_keywords(keywords)
        ElementParameterValidator.validate_duration(duration)

        images = []
        if duration > MAX_DURATION_PER_IMAGE:
            number_of_images = int(duration / MAX_DURATION_PER_IMAGE)
            for i in range(number_of_images):
                images.append({
                    'keywords': keywords,
                    'duration': MAX_DURATION_PER_IMAGE
                })

            remaining_duration = duration % MAX_DURATION_PER_IMAGE
            if remaining_duration <= MIN_DURATION_PER_IMAGE:
                # We won't create an image animation that is too short, we
                # will make the previous one longer
                images[len(images) - 1]['duration'] += remaining_duration
            else:
                images.append({
                    'keywords': keywords,
                    'duration': remaining_duration
                })
        else:
            images.append({
                'keywords': keywords,
                'duration': duration
            })

    @classmethod
    def build_from_segment(cls, segment: dict):
        keywords = segment.get(SegmentField.KEYWORDS.value, None)
        # TODO: I should always have 'calculated_duration' when duration
        # has been processed
        duration = segment.get('calculated_duration', None)

        cls.build(keywords, duration)

    @classmethod
    def build(cls, keywords: str, duration: Union[float, int]):
        images = cls.get_images_array(keywords, duration)

        videos = []
        for image in images:
            videos.append(cls.ai_image_to_video(image.keywords, image.duration, None))

        video = concatenate_videoclips(videos)

        return video
    
    @classmethod
    def ai_image_to_video(cls, keywords: str, duration: Union[float, int], effect = None):
        """
        Creates an AI image with the provided 'keywords' and builds a video
        that lasts 'duration' seconds with the given 'effect'.

        TODO: Apply 'effect' that is not working by now
        """
        ElementParameterValidator.validate_keywords(keywords)
        ElementParameterValidator.validate_duration(duration)
        
        return ImageElementBuilder.image_to_video(create_ai_image(keywords), duration, effect)


from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.enums import SegmentType
from youtube_autonomous.elements.validator import BUILDER_SUBCLASSES
from youtube_autonomous.segments.builder.ai import create_ai_narration
from typing import Union


class ElementBuilder:
    @staticmethod
    def get_subclasses():
        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        return [
            AIImageElementBuilder,
            AIVideoElementBuilder,
            ImageElementBuilder,
            VideoElementBuilder,
            CustomStockElementBuilder,
            StockElementBuilder,
            MemeElementBuilder,
            SoundElementBuilder,
            YoutubeVideoElementBuilder,
            TextElementBuilder,
            PremadeElementBuilder,
            EffectElementBuilder,
            GreenscreenElementBuilder
        ]

    @staticmethod
    def get_subclasses_as_str():
        return ', '.join(ElementBuilder.get_subclasses())
    
    @staticmethod
    def get_subclass_by_type(type: Union[SegmentType, str]):
        type = ElementParameterValidator.validate_segment_type(type)

        from youtube_autonomous.elements.builder.ai_image_element_builder import AIImageElementBuilder
        from youtube_autonomous.elements.builder.image_element_builder import ImageElementBuilder
        from youtube_autonomous.elements.builder.ai_video_element_builder import AIVideoElementBuilder
        from youtube_autonomous.elements.builder.video_element_builder import VideoElementBuilder
        from youtube_autonomous.elements.builder.custom_stock_element_builder import CustomStockElementBuilder
        from youtube_autonomous.elements.builder.stock_element_builder import StockElementBuilder
        from youtube_autonomous.elements.builder.meme_element_builder import MemeElementBuilder
        from youtube_autonomous.elements.builder.sound_element_builder import SoundElementBuilder
        from youtube_autonomous.elements.builder.youtube_video_element_builder import YoutubeVideoElementBuilder
        from youtube_autonomous.elements.builder.text_element_builder import TextElementBuilder
        from youtube_autonomous.elements.builder.premade_element_builder import PremadeElementBuilder
        from youtube_autonomous.elements.builder.effect_element_builder import EffectElementBuilder
        from youtube_autonomous.elements.builder.greenscreen_element_builder import GreenscreenElementBuilder

        if type == SegmentType.MEME:
            return MemeElementBuilder
        elif type == SegmentType.AI_IMAGE:
            return AIImageElementBuilder
        elif type == SegmentType.AI_VIDEO:
            return AIVideoElementBuilder
        elif type == SegmentType.IMAGE:
            return ImageElementBuilder
        elif type == SegmentType.VIDEO:
            return VideoElementBuilder
        elif type == SegmentType.STOCK:
            return StockElementBuilder
        elif type == SegmentType.CUSTOM_STOCK:
            return CustomStockElementBuilder
        elif type == SegmentType.SOUND:
            return SoundElementBuilder
        elif type == SegmentType.YOUTUBE_VIDEO:
            return YoutubeVideoElementBuilder
        elif type == SegmentType.TEXT:
            return TextElementBuilder
        elif type == SegmentType.PREMADE:
            return PremadeElementBuilder
        elif type == SegmentType.EFFECT:
            return EffectElementBuilder
        elif type == SegmentType.GREENSCREEN:
            return GreenscreenElementBuilder
    
    @classmethod
    def build_narration(cls, text: str, output_filename: str):
        """
        Generates a narration file that narrates the 'text' provided and
        is stored locally as 'output_filename'. If 'text' or 
        'output_filename' fields are not provided it will raise an 
        Exception.
        """
        ElementParameterValidator.validate_string_mandatory_parameter(text)
        ElementParameterValidator.validate_string_mandatory_parameter(output_filename)

        return create_ai_narration(text, output_filename = output_filename)
    
    @classmethod
    def handle_narration_from_segment(cls, segment: dict):
        pass


FILE_DURATION = 99999
"""
Valid int duration for specific shortcodes and enhance elements
that need to calculate the real duration in the last moment
when the resource is available because the duration will be that
resource duration.
"""
SEGMENT_DURATION = 999999
"""
Valid int duration for specific enhancement elements that need to
calculate the real duration in the last moment when the whole
segment video is available because the duration will be that 
segment duration.
"""


from youtube_autonomous.segments.enhancement.edition_manual.enums import EditionManualTermMode, EditionManualTermContext
from youtube_autonomous.segments.enums import EnhancementElementStart, EnhancementElementDuration
from youtube_autonomous.segments.enhancement.enhancement_element import EnhancementElement
from typing import Union


class EditionManualTerm:
    """
    A term of the edition manual terms book that is used to look
    for matches in the segments text in order to enhance the video
    or audio content.

    See: https://www.notion.so/Diccionarios-de-mejora-155efcba8f0d44e0890b178effb3be84?pvs=4
    """
    @property
    def term(self):
        # TODO: Explain the term
        return self._term
    
    @term.setter
    def term(self, term: str):
        if not term:
            raise Exception('no "term" provided.')
        
        if not isinstance(term, str):
            raise Exception('The "term" parameter provided is not a string.')
        
        self._term = term

    @property
    def mode(self):
        # TODO: Explain the term
        return self._mode
        
    @mode.setter
    def mode(self, mode: Union[EditionManualTermMode, str]):
        if not mode:
            raise Exception('No "mode" provided.')
        
        if not isinstance(mode, (EditionManualTermMode, str)):
            raise Exception('The "mode" parameter provided is not a valid EditionManualTermMode nor a string.')
        
        if isinstance(mode, str):
            if not EditionManualTermMode.is_valid(mode):
                raise Exception('The "mode" parameter provided is not a valid EditionManualTermMode value.')
            
            mode = EditionManualTermMode(mode)

        self._mode = mode

    @property
    def context(self):
        # TODO: Explain the term
        return self._context
    
    @context.setter
    def context(self, context: Union[EditionManualTermContext, str]):
        if not context:
            raise Exception('No "context" provided.')
        
        if not isinstance(context, (EditionManualTermContext, str)):
            raise Exception('The "context" parameter provided is not a valid EditionManualTermContext nor a string.')
        
        if isinstance(context, str):
            if not EditionManualTermContext.is_valid(context):
                raise Exception('The "context" parameter provided is not a valid EditionManualTermContext value.')
            
            context = EditionManualTermContext(context)

        self._context = context

    @property
    def enhancements(self):
        return self._enhancements
    
    @enhancements.setter
    def enhancements(self, enhancements: list[EnhancementElement, dict]):
        # TODO: Make some checkings and improvements
        if not enhancements:
            enhancements = []

        # We turn dicts to EnhancementElement if necessary
        obj_enhancements = []
        for enhancement in enhancements:
            if not isinstance(enhancement, EnhancementElement) and not issubclass(enhancement, EnhancementElement):
                obj_enhancements.append(EnhancementElement.get_class_from_type(enhancement['type'])(enhancement['type'], EnhancementElementStart.START_OF_FIRST_SHORTCODE_CONTENT_WORD, EnhancementElementDuration.SHORTCODE_CONTENT, enhancement['keywords'], enhancement.get('url', ''), enhancement.get('filename', ''), enhancement['mode']))
            else:
                obj_enhancements.append(enhancement)

        self._enhancements = obj_enhancements

    def __init__(self, term: str, mode: str, context: str, enhancements: list[dict]):
        self.term = term
        self.mode = mode
        self.context = context
        self.enhancements = enhancements


from yta_general_utils.programming.enum import YTAEnum as Enum, get_values


class SegmentType(Enum):
    """
    These Enums represent the types that a Segment could be, allowing
    us to check and detect if it is valid.
    """
    # Interesting: https://docs.python.org/3/howto/enum.html
    CUSTOM_STOCK = 'custom_stock'
    """
    Stock videos but extracted from our own custom sources.
    """
    STOCK = 'stock'
    """
    Stock videos extracted from external stock platforms.
    """
    AI_IMAGE = 'ai_image'
    IMAGE = 'image'
    AI_VIDEO = 'ai_video'
    VIDEO = 'video'
    SOUND = 'sound'
    YOUTUBE_VIDEO = 'youtube_video'
    TEXT = 'text'
    MEME = 'meme'
    EFFECT = 'effect'
    PREMADE = 'premade'
    GREENSCREEN = 'greenscreen'

    # TODO: This below is now available through the rules...
    # This should be removed in a near future
    @classmethod
    def get_premade_types(cls):
        """
        Returns a list containing all the Segment Types that are 
        premades.
        """
        return [
            # SegmentType.YOUTUBE_SEARCH,
            # SegmentType.GOOGLE_SEARCH
        ]

    @classmethod
    def get_narration_types(cls):
        """
        Returns the SegmentType enums that are compatible with 
        audio narration.
        """
        return cls.get_all()
    
    @classmethod
    def get_narration_type_values(cls):
        """
        Returns the SegmentType enums values that are compatible
        with audio narration.
        """
        return get_values(cls.get_narration_types())
    
    @classmethod
    def get_url_types(cls):
        """
        Returns the SegmentType enums that are compatible with the
        'url' parameter.
        """
        return [
            SegmentType.IMAGE,
            SegmentType.YOUTUBE_VIDEO,
            SegmentType.VIDEO,
            SegmentType.SOUND
        ]
    
    @classmethod
    def get_url_type_values(cls):
        """
        Returns the SegmentType enums values that are compatible
        with the 'url' parameter.
        """
        return get_values(cls.get_url_types())
    
    @classmethod
    def get_keywords_types(cls):
        """
        Returns the SegmentType enums that are compatible with the
        'keywords' parameter.
        """
        return [
            SegmentType.MEME,
            SegmentType.AI_IMAGE,
            SegmentType.CUSTOM_STOCK,
            SegmentType.STOCK,
            # TODO: Add IMAGE in the future with Bing or Google Search
            # TODO: Add YOUTUBE_VIDEO in the future for Youtube Search
        ]
    
    @classmethod
    def get_keywords_type_values(cls):
        """
        Returns the SegmentType enums values that are compatible 
        with the 'keywords' parameter.
        """
        return get_values(cls.get_keywords_types())
    
    @classmethod
    def get_filename_types(cls):
        """
        Returns the SegmentType enums that are compatible with the
        'filename' parameter.
        """
        return [
            SegmentType.IMAGE,
            SegmentType.SOUND,
            SegmentType.VIDEO
        ]
    
    @classmethod
    def get_filename_type_values(cls):
        """
        Returns the SegmentType enums values that are compatible 
        with the 'filename' parameter.
        """
        return get_values(cls.get_filename_types())
    
    @classmethod
    def get_text_types(cls):
        """
        Returns the SegmentType enums that are compatible with the
        'text' parameter.
        """
        # TODO: This is defined in element rules
        return [
            SegmentType.TEXT,
        ]
    
    @classmethod
    def get_text_type_values(cls):
        """
        Returns the SegmentType enums values that are compatible 
        with the 'text' parameter.
        """
        return get_values(cls.get_text_types())
    
class SegmentField(Enum):
    """
    These Enums represent the fields that a Segment has, allowing us
    to check that any required field is provided and/or to detect 
    which one is missing.

    Examples: TYPE, KEYWORDS, URL, etc.
    """
    # Interesting: https://docs.python.org/3/howto/enum.html
    TYPE = 'type'
    KEYWORDS = 'keywords'
    URL = 'url'
    FILENAME = 'filename'
    NARRATION_TEXT = 'narration_text'
    VOICE = 'voice'
    TEXT = 'text'
    DURATION = 'duration'
    AUDIO_NARRATION_FILENAME = 'audio_narration_filename'
    MUSIC = 'music'

class EnhancementField(Enum):
    """
    Fields accepted for enhancement elements.
    """
    TYPE = 'type'
    KEYWORDS = 'keywords'
    URL = 'url'
    FILENAME = 'filename'
    NARRATION_TEXT = 'narration_text'
    VOICE = 'voice'
    TEXT = 'text'
    AUDION_NARRATION_FILENAME = 'audio_narration_filename'
    MUSIC = 'music'
    START = 'start'
    DURATION = 'duration'
    
class SegmentBuildingField(Enum):
    """
    The fields that are used when building the segment and are
    not provided by the user in the initial segment json data.
    """
    TRANSCRIPTION = 'transcription'
    AUDIO_FILENAME = 'audio_filename'
    AUDIO_CLIP = 'audio_clip'
    VIDEO_FILENAME = 'video_filename'
    VIDEO_CLIP = 'video_clip'
    FULL_FILENAME = 'full_filename'
    FULL_CLIP = 'full_clip'
    # TODO: What about 'status', 'music_filename' and any other (if existing) (?)

class ProjectField(Enum):
    """
    The fields that are used to handle the project information.
    """
    STATUS = 'status'
    """
    The current status of the project, that must be a ProjectStatus
    enum.
    """
    SEGMENTS = 'segments'
    """
    The array that contains all the information about each one of 
    his segments.
    """

class EnhancementElementType(Enum):
    """
    These Enums represent the type of elements we can use to enhance
    a segment.
    """
    MEME = 'meme'
    STICKER = 'sticker'
    IMAGE = 'image'
    SOUND = 'sound'
    GREENSCREEN = 'greenscreen'
    EFFECT = 'effect'
    # TODO: Add more in the future please 'video', 'ai_image', etc.
    # maybe all the segment types
    """
    This enum must be used only for validation process. This has been made to 
    check if some provided type is valid as it is one of the valid and 
    available ones.
    """

class EnhancementElementField(Enum):
    """
    These Enums represent the type of fields we can use with the
    segment enhancement elements to fit correctly the segment.
    """
    # IS_ACTIVE = 'is_active'
    # """
    # This field determines if the current enhancement element is active or not,
    # allowing us to have the information but applying not this element for an
    # specific segment.
    # """
    START = 'start'
    """
    This field determines the time moment (related to the segment video time)
    in which the enhancement element should start, expressed in seconds. This
    can be a number or an EnhancementElementStart enum.
    """
    DURATION = 'duration'
    """
    This field determines the time that the enhancement element should last 
    since the 'start' moment, expressed in seconds. This can be a number or
    an EnhancementElementDuration enum.
    """
    KEYWORDS = 'keywords'
    FILENAME = 'filename'
    URL = 'url'
    MODE = 'mode'
    
    # TODO: In the future, let this be more personalized
    # NARRATION_TEXT = 'narration_text'
    # VOICE = 'voice'
    # TEXT = 'text'
    # AUDIO_NARRATION_FILENAME = 'audio_narration_filename'
    # MUSIC = 'music'
    #ORIGIN = 'origin'

class EnhancementElementStart(Enum):
    """
    These Enums are valid values for the 'start' EnhancementElementField.
    """
    BETWEEN_WORDS = 'between_words'
    """
    This will make the enhancement element start just in the middle of 
    two words that are dictated in narration. This means, after the end
    of the first and and before the start of the next one (that should
    fit a silence part).
    """
    START_OF_FIRST_SHORTCODE_CONTENT_WORD = 'start_of_first_shortcode_content_word'
    """
    This will make the enhancement element start when the first word of 
    the shortcode content starts being dictated.
    """
    MIDDLE_OF_FIRST_SHORTCODE_CONTENT_WORD = 'middle_of_first_shortcode_content_word'
    """
    This will make the enhancement element start when the first word of 
    the shortcode content is in the middle of the dictation.
    """
    END_OF_FIRST_SHORTCODE_CONTENT_WORD = 'end_of_first_shortcode_content_word'
    """
    This will make the enhancement element start when the first word of 
    the shortcode content ends being dictated.
    """

    @classmethod
    def get_default(cls):
        return cls.START_OF_FIRST_SHORTCODE_CONTENT_WORD
    
class EnhancementElementDuration(Enum):
    """
    These Enums represent the time, during a segment lifetime, that
    the elements are going to last (be shown).
    """
    SHORTCODE_CONTENT = 'shortcode_content'
    """
    This will make the enhancement element last until the shortcode
    block-scoped content is narrated.
    """
    FILE_DURATION = 'file_duration'
    """
    This will make the segment last the clip duration. It will be
    considered when the file is downloaded, and that duration will
    be flagged as 9999. This is for videos or audios that have a
    duration based on file.
    """
    
    @classmethod
    def get_default(cls):
        return cls.SHORTCODE_CONTENT

class EnhancementElementMode(Enum):
    """
    These Enums represent the different ways in which the project
    segment elements can be built according to the way they are
    included in the segment.
    """
    INLINE = 'inline'
    """
    Those segment elements that will be displayed in 'inline' mode, that
    means they will interrupt the main video, be played, and then go back
    to the main video. This will modify the clip length, so we need to 
    refresh the other objects start times.
    """
    OVERLAY = 'overlay'
    """
    Those segment elements that will be displayed in 'overlay' mode, that
    means they will be shown in the foreground of the main clip, changing
    not the main video duration, so they don't force to do any refresh.
    """
    REPLACE = 'replace'
    """
    Those enhancement elements that will replace the video in this mode.
    This means that the original video is modified by this enhancement
    element and that modified part will be placed instead of the original
    video. This modified part could be the whole video or only a part of
    it. This is how most of the greenscreens or effects are applied.
    """
    
    @classmethod
    def get_default(cls):
        """
        Returns the default enum of this list. This value will be used when
        no valid value is found.
        """
        return cls.INLINE

class SegmentConstants(Enum):
    """
    These Enums are constants we need to use during the project
    and segments processing.
    """
    FILE_DURATION = 9999
    """
    This is the duration, as int value, that will be considered as
    the original video duration and will be used to set the original
    clip (that can be downloaded or load from local storage) duration
    as the final duration. This amount is just to be recognized and 
    will be replaced, when file processed, by the real duration.
    """

class SegmentElementOrigin(Enum): 
    """
    These Enums represent the way in which the project segment elements
    are searched and obtained from.
    """   
    YOUTUBE = 'youtube'
    """
    Those segment elements that will be searched in Youtube and downloaded
    if found and available.
    """
    FILE = 'file'
    """
    Those segment elements that will be load from local storage. The 
    location must be set in the segment building object.
    """
    DEFAULT = YOUTUBE
    """
    This is the default value I want (me, the developer, is the one who
    sets it by now). This one will be pointing to another of the available
    Enums. This Enum has been created to make it easy building objects in
    code and having the default value in only one strict place.
    """
    ALL = [YOUTUBE, FILE]
    """
    This is for validation purposes only.
    """

class SegmentStatus(Enum):
    """
    The current segment status defined by this string.
    """
    TO_START = 'to_start'
    IN_PROGRESS = 'in_progress'
    FINISHED = 'finished'

class ProjectStatus(Enum):
    """
    The current project status defined by this string.
    """
    TO_START = 'to_start'
    IN_PROGRESS = 'in_progress'
    FINISHED = 'finished'


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.segments.enums import SegmentField
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from youtube_autonomous.segments.builder.youtube.youtube_downloader import YoutubeDownloader
from moviepy.editor import AudioFileClip
from typing import Union


class SoundElementBuilder(ElementBuilder):
    @classmethod
    def build_from_segment(cls, segment: dict):
        keywords = segment.get(SegmentField.KEYWORDS.value, None)
        # TODO: I should always have 'calculated_duration' when duration
        # has been processed
        duration = segment.get('calculated_duration', None)

        return cls.build(keywords, duration)

    @classmethod
    def build(cls, keywords: str, duration: Union[float, int]) -> AudioFileClip:
        ElementParameterValidator.validate_keywords(keywords)
        ElementParameterValidator.validate_duration(duration)

        youtube_downloader = YoutubeDownloader()

        youtube_downloader.deactivate_ignore_repeated()
        temp_filename = youtube_downloader.download_sound_audio(keywords)
        youtube_downloader.activate_ignore_repeated()

        # TODO: Look for a better strategy (?)
        if not temp_filename:
            raise Exception('No sound found with the given "keywords": ' + str(keywords) + '.')
        
        audio = AudioFileClip(temp_filename)

        # TODO: I apply the duration but I don't change it
        if duration < audio.duration:
            audio = audio.subclip(0, duration)



from youtube_autonomous.elements.validator.error_message import ErrorMessage
from youtube_autonomous.segments.enums import SegmentType
from youtube_autonomous.elements.validator import RULES_SUBCLASSES, BUILDER_SUBCLASSES
from yta_general_utils.programming.parameter_validator import ParameterValidator
from yta_general_utils.file.filename import FileType
from typing import Union


class ElementParameterValidator(ParameterValidator):
    """
    A validator that will raise an Exception if any of the used
    checkings is invalid, or will return the value if ok.
    """
    @classmethod
    def validate_keywords(cls, keywords: str):
        cls.validate_string_mandatory_parameter('keywords', keywords)

        return keywords

    @classmethod
    def validate_text(cls, text: str):
        cls.validate_string_mandatory_parameter('text', text)

        return text
    
    @classmethod
    def validate_premade_name(cls, premade_name: str):
        cls.validate_string_mandatory_parameter('premade_name', premade_name)
        # TODO: Maybe validate that is a key of a valid premade (?)

        return premade_name
    
    @classmethod
    def validate_text_class_name(cls, text_class_name: str):
        cls.validate_string_mandatory_parameter('text_class_name', text_class_name)
        # TODO: Maybe validate that is a key of a valid text_class (?)

        return text_class_name
    
    @classmethod
    def validate_duration(cls, duration: Union[int, float]):
        # TODO: Maybe validate that is a key of a valid text class (?)
        cls.validate_numeric_positive_mandatory_parameter('duration', duration)

        return duration
    
    @classmethod
    def validate_url(cls, url: str, is_mandatory: True):
        """
        A mandatory 'url' must be not None, a valid string and also
        a valid url and accessible.

        A non mandatory 'url' can be None, 
        """
        if not is_mandatory and not url:
            return True
        
        cls.validate_mandatory_parameter('url', url)
        cls.validate_string_parameter('url', url)
        cls.validate_url_is_ok(url)
        # TODO: Validate 'url' file is some FileType

        return url

    @classmethod
    def validate_filename(cls, filename: str, file_type: FileType = FileType.IMAGE, is_mandatory: bool = True):
        """
        Validates the 'filename' parameter. If the 'filename' parameter has
        some value, all conditions will be checked. If there is not 'filename'
        and 'is_mandatory' is False, no Exceptions will be raised because it
        is valid.
        """
        if not is_mandatory and not filename:
            return True
        
        cls.validate_mandatory_parameter('filename', filename)
        cls.validate_string_parameter('filename', filename)
        cls.validate_file_exists('filename', filename)
        cls.validate_filename_is_type('filename', filename, file_type)
        # TODO: Validate type trying to instantiate it

        return filename

    @classmethod
    def validate_rules(cls, rules: 'ElementRules'):
        if not type(rules).__name__ in RULES_SUBCLASSES:
            raise Exception(ErrorMessage.parameter_is_not_rules('rules'))
        
        return rules
        
    @classmethod
    def validate_builder(cls, builder: 'ElementBuilder'):
        if not type(builder).__name__ in BUILDER_SUBCLASSES:
            raise Exception(ErrorMessage.parameter_is_not_builder('builder'))
        
        return builder
        
    @classmethod
    def validate_segment_type(cls, type: Union[SegmentType, str]):
        cls.validate_mandatory_parameter('type', type)
        cls.validate_is_class('type', type, ['SegmentType', 'str'])

        if isinstance(type, str):
            if not SegmentType.is_valid(type):
                raise Exception('The "type" parameter provided is not a valid SegmentType value.')
            
            type = SegmentType(type)

        return type


from youtube_autonomous.segments.enums import SegmentField
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator


class RulesChecker:
    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, rules: 'ElementRules'):
        ElementParameterValidator.validate_rules(rules)
        
        self._rules = rules

    def __init__(self, rules: 'ElementRules'):
        self.rules = rules

    def check_this_need_rules(self, element: dict):
        """
        Receives the 'element' dictionary that must contain the fields that
        are mandatory by the rules configuration. This method will raise an
        Exception if any rule is broken.

        If no Exception is raised, all the needed fields are provided. This
        is not verifying if possible fields are available or not. Narration
        can be enabled but not needed, and maybe only one 'narration_text'
        is provided and not 'voice'. This situation won't raise an Exception
        here if narration is not needed.

        This method must be used when loading the element to raise any
        Exception if something is wrong and interrumpting the execution until
        it is fixed.
        """
        return RulesChecker.check_need_rules(element, self.rules)
    
    # With this methods below we check if we should build an specific
    # part. This must be used after validating the needed rules to
    # raise any exception if something is missing and is mandatory.
        
    def should_build_narration_rule(self, element: dict):
        """
        Returns True if, according to the rules given in the initialization,
        the element must have or can have a narration and the needed fields
        are given. This guarantees that the narration can be done and must 
        be done.
        """
        return (self.rules.can_have_narration or self.rules.need_narration) and RulesChecker.is_need_narration_rule_satisfied(element)

    def should_build_specific_duration_rule(self, element: dict):
        """
        Returns True if, according to the rules given in the initialization,
        the element must have or can have an specific duration and that field
        is given. This guarantees that the element must have the specific
        'duration' set during the building process.
        """
        return (self.rules.can_have_specific_duration or self.rules.need_specific_duration) and RulesChecker.is_need_specific_duration_rule_satisfied(element)

    def should_build_text_rule(self, element: dict):
        """
        Returns True if, according to the rules given in the initialization,
        the element must have or can have the text and that field is given.
        This guarantees that the element can use that text for the building
        process.
        """
        return (self.rules.can_have_text or self.rules.need_text) and RulesChecker.is_need_text_rule_satisfied(element)

    def should_build_filename_or_url_rule(self, element: dict):
        """
        Returns True if, according to the rules given in the initialization,
        the element must have or can have the filename or the url and those
        fields are given. This guarantees that the element can use those 
        fields, according to our strategy, in the building process.
        """
        # TODO: This is not ok
        return (self.rules.can_have_url or self.rules.can_have_filename or self.rules.need_filename_or_url) and RulesChecker.is_need_filename_or_url_rule_satisfied(element)

    def should_build_keywords_rule(self, element: dict):
        """
        Returns True if, according to the rules given in the initialization,
        the element must have or can have the keywords and that field is 
        given. This guarantees that the element can use that field in the
        building process.
        """
        return (self.rules.can_have_keywords or self.rules.need_keywords) and RulesChecker.is_need_keywords_rule_satisfied(element)

    @classmethod
    def check_need_rules(cls, element: dict, rules: 'ElementRules'):
        """
        Receives the 'element' dictionary that must contain the fields that
        are mandatory by the rules configuration. This method will raise an
        Exception if any rule is broken.

        If no Exception is raised, all the needed fields are provided. This
        is not verifying if possible fields are available or not. Narration
        can be enabled but not needed, and maybe only one 'narration_text'
        is provided and not 'voice'. This situation won't raise an Exception
        here if narration is not needed.

        This method must be used when loading the element to raise any
        Exception if something is wrong and interrumpting the execution until
        it is fixed.
        """
        ElementParameterValidator.validate_mandatory_parameter('element', element)
        ElementParameterValidator.validate_mandatory_parameter('rules', rules)
        ElementParameterValidator.validate_rules(rules)
        
        if rules.need_narration and not cls.is_need_narration_rule_satisfied(element):
            raise Exception('Narration is needed for this element and necessary fields are not present. The "audio_narration_filename" field or "voice" and "narration_text" fields are needed.')
        
        if rules.need_specific_duration and not cls.is_need_specific_duration_rule_satisfied(element):
            raise Exception('Specific duration is needed for this element and necessary fields are not present. The "duration" field is needed, or a narration with "audio_narration_filename" field or "voice" and "narration_text" fields.')
        
        if rules.need_text and not cls.is_need_text_rule_satisfied(element):
            raise Exception('An specific "text" field is needed and does not exist.')
        
        if rules.need_filename_or_url and not cls.is_need_filename_or_url_rule_satisfied(element):
            raise Exception('The "filename" field or the "url" field is needed (at least one of them) and none is provided.')
        
        if rules.need_keywords and not cls.is_need_keywords_rule_satisfied(element):
            raise Exception('The "keywords" field is needed for this type.')
        
    # These methods below check if the needed fields of the different
    # need rules exist so it is possible to build the corresponding
    # part, returning True if yes or False if not.

    @classmethod
    def is_need_narration_rule_satisfied(cls, element: dict):
        """
        Checks if the 'need_narration' rule is satisfied or not. This
        will check if the 'audio_narration_filename' or the tuple 'voice'
        and 'narration_text' are given.
        """
        narration_text = element.get(SegmentField.NARRATION_TEXT.value, None)
        voice = element.get(SegmentField.VOICE.value, None)
        audio_narration_filename = element.get(SegmentField.AUDIO_NARRATION_FILENAME.value, None)

        return audio_narration_filename or (voice and narration_text)

    @classmethod
    def is_need_specific_duration_rule_satisfied(cls, element: dict):
        """
        Checks if the 'need_specific_duration' rule is satisfied or not.
        This will check if the specific 'duration' field is set or if it
        is possible to create the narration (the 'need_narration' rule is
        satisfied).
        """
        duration = element.get(SegmentField.DURATION.value, None)

        return duration or cls.is_need_narration_rule_satisfied(element)

    @classmethod
    def is_need_text_rule_satisfied(cls, element: dict):
        """
        Checks if the 'need_text' rule is satisfied by checking if the
        'text' field is given or not.
        """
        text = element.get(SegmentField.TEXT.value, None)

        return text

    @classmethod
    def is_need_filename_or_url_rule_satisfied(cls, element: dict):
        """
        Checks if the 'need_filename_or_url' rule is satisfied by checking
        if at least one of the tuple 'filename' and 'url' fields are 
        provided or not.        
        """
        filename = element.get(SegmentField.FILENAME.value, None)
        url = element.get(SegmentField.URL.value, None)

        return filename or url

    @classmethod
    def is_need_keywords_rule_satisfied(cls, element: dict):
        """
        Checks if the 'need_keywords' rule is satisfied by checking if
        the 'keywords' field is provided.
        """
        keywords = element.get(SegmentField.KEYWORDS.value, None)

        return keywords


from youtube_autonomous.segments.enums import SegmentType, SegmentField, EnhancementElementDuration
from youtube_autonomous.elements.meme_element import MemeElement
from youtube_autonomous.elements.element_data import ElementData
from youtube_autonomous.database.database_handler import DatabaseHandler
from youtube_autonomous.shortcodes.shortcode_parser import ShortcodeParser
from youtube_autonomous.segments.builder.ai import create_ai_narration
from youtube_autonomous.shortcodes.objects.shortcode import Shortcode
from yta_general_utils.logger import print_in_progress, print_completed
from yta_general_utils.file import copy_file
from yta_general_utils.file.filename import get_file_extension
from moviepy.editor import VideoFileClip, AudioFileClip
from bson.objectid import ObjectId
from typing import Union


class Segment:
    @property
    def status(self):
        return self._status
    
    @status.setter
    def status(self, status: str):
        self._status = status
        if self._do_update_database:
            self.database_handler.update_project_segment_status(self.project_id, self.index, status)

    @property
    def audio_filename(self):
        audio_filename = self.data.audio_filename

        if audio_filename and not self.audio_clip:
            self.audio_clip = AudioFileClip(audio_filename)

        return audio_filename
    
    @audio_filename.setter
    def audio_filename(self, audio_filename: Union[str, None]):
        self.data.audio_filename = audio_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, SegmentField.AUDIO_FILENAME.value, audio_filename)

        if self.data.audio_filename:
            self.audio_clip = AudioFileClip(self.data.audio_filename)

    @property
    def video_filename(self):
        video_filename = self.data.video_filename

        if video_filename and not self.video_clip:
            self.video_clip = VideoFileClip(video_filename)

        return video_filename
    
    @video_filename.setter
    def video_filename(self, video_filename: Union[str, None]):
        self.data.video_filename = video_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, SegmentField.VIDEO_FILENAME.value, video_filename)

        if self.data.video_filename:
            self.video_clip = VideoFileClip(self.data.video_filename)

    @property
    def full_filename(self):
        full_filename = self.data.full_filename

        if full_filename and not self.full_clip:
            self.full_clip = VideoFileClip(full_filename)

        return full_filename
    
    @full_filename.setter
    def full_filename(self, full_filename: Union[str, None]):
        self.data.full_filename = full_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, SegmentField.FULL_FILENAME.value, full_filename)

        if self.data.full_filename:
            self.full_clip = VideoFileClip(self.data.full_filename)

    @property
    def audio_narration_filename(self):
        return self.data.audio_narration_filename
    
    @audio_narration_filename.setter
    def audio_narration_filename(self, audio_narration_filename: Union[str, None]):
        self.data.audio_narration_filename = audio_narration_filename

        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, SegmentField.AUDIO_NARRATION_FILENAME.value, audio_narration_filename)

    @property
    def transcription(self):
        return self.data.transcription
    
    @transcription.setter
    def transcription(self, transcription: Union[dict, None]):
        self.data.transcription = transcription
        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, 'transcription', transcription)

    @property
    def shortcodes(self):
        return self.data.shortcodes
    
    @shortcodes.setter
    def shortcodes(self, shortcodes: Union[list[Shortcode], None]):
        self.data.shortcodes = shortcodes
        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, 'shortcodes', shortcodes)

    @property
    def calculated_duration(self):
        return self.data.calculated_duration
    
    @calculated_duration.setter
    def calculated_duration(self, calculated_duration: Union[float, int, None]):
        self.data.calculated_duration = calculated_duration
        if self._do_update_database:
            self.database_handler.update_project_segment_field(self.project_id, self.index, 'calculated_duration', calculated_duration)

    @property
    def text(self):
        return self.data.text
    
    @property
    def narration_text(self):
        return self.data.narration_text

    def __init__(self, project_id: Union[str, ObjectId], index: int, type: Union[str, SegmentType], status: str, data: dict):
        if not type:
            raise Exception('No "type" provided.')
        
        if isinstance(type, str):
            if not SegmentType.is_valid(type):
                raise Exception('The "type" provided is not a valid SegmentType.')
            
            type = SegmentType(type)

        # TODO: Check 'data'

        self._do_update_database = False

        self.type = type
        self.project_id = str(project_id)
        self.index = index
        self.status = status
        
        if self.type == SegmentType.MEME:
            self.element = MemeElement()
            self.data = ElementData(data)

        self.database_handler = DatabaseHandler()
        self.shortcode_parser = ShortcodeParser()

        self.effects = data.get('effects', [])

        self._do_update_database = True

    def validate(self):
        # 1. If 'text' have shortcodes
        self.shortcode_parser.parse(self.text)

        # 2. If 'narration_text' have invalid shortcodes
        self.shortcode_parser.parse(self.narration_text)

        # 3. If 'data.effects' are invalid
        from youtube_autonomous.elements.rules.effect_element_rules import EffectElementRules
        from youtube_autonomous.elements.rules.rules_checker import RulesChecker
        for effect in self.effects:
            RulesChecker(EffectElementRules()).check_this_need_rules(effect)
        # TODO: I need to check any other modification I let them
        # use, such as Greenscreens, etc.

        # 4. If element has not necessary parameters
        self.element.rules_checker.check_this_need_rules(self.data)

    def prepare(self):
        # Check duration
        # TODO: This has to be done when storing the project
        if self.data.duration and not self.calculated_duration:
            if isinstance(self.data.duration, str):
                # TODO: Make this consts and/or Enum
                if self.data.duration == 'SHORTCODE_CONTENT':
                    self.data.duration = 99997
                elif self.data.duration == 'FILE_DURATION':
                    self.data.duration = 99998
                elif self.data.duration == 'SEGMENT_DURATION':
                    self.data.duration = 99999
            self.calculated_duration = self.data.duration

        # Parse narration text and remove shortcodes
        # TODO: This has to be done when storing the project
        if self.element.rules_checker.should_build_narration_rule(self.data) and not self.data.narration_text_sanitized_without_shortcodes:
            self.shortcode_parser.parse(self.data.narration_text)
            self.data.narration_text_sanitized_without_shortcodes = self.shortcode_parser.text_sanitized_without_shortcodes
            # TODO: I think these 2 below are useless
            self.data.narration_text_with_simplified_shortcodes = self.shortcode_parser.text_sanitized_with_simplified_shortcodes
            self.data.narration_text_sanitized = self.shortcode_parser.text_sanitized

        # Generate narration audio
        if self.element.rules_checker.should_build_narration_rule(self.data) and not self.audio_clip:
            self.handle_narration()

        # Generate video
        if not self.video_clip:
            self.video_clip = self._builder.build_segment(self)
            # We write this video part to be able to recover it in the future
            segment_part_filename = self.__create_segment_file('video.mp4')
            self.video_clip.write_videofile(segment_part_filename)
            self.video_filename = segment_part_filename
            print_completed('Created basic video part and stored locally and in database.')


        



    def handle_narration(self):
        if self.audio_narration_filename:
            segment_part_filename = self.__create_segment_file(f'narration.{get_file_extension(self.audio_narration_filename)}')
            copy_file(self.audio_narration_filename, segment_part_filename)
            print_completed('Original voice narration file copied to segment parts folder')
            self.audio_narration_filename = segment_part_filename
            self.audio_filename = segment_part_filename
        else:
            segment_part_filename = self.__create_segment_file('narration.wav')
            # TODO: Voice parameter need to change
            self.audio_filename = create_ai_narration(self.data.narration_text_sanitized_without_shortcodes, output_filename = segment_part_filename)
            print_completed('Voice narration created successfully')

    def build(self):
        pass


    def __create_segment_file(self, filename: str):
        """
        Creates a filename within the definitive segments folder
        to keep the generated file locally to recover it in the
        next project execution if something goes wrong. The 
        definitive filename will be built using the provided
        'filename' and adding some more information in the name.

        This method will generate a temporary filename that uses
        the current segment index in its name and is placed in 
        segment parts folder.

        This method returns the final filename created.
        """
        if not filename:
            raise Exception('No "filename" provided.')

        temp_filename = get_temp_filename(filename)

        return f'{DEFAULT_SEGMENT_PARTS_FOLDER}/segment_{self.segment_index}_{temp_filename}'


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.segments.enums import SegmentType
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator
from yta_multimedia.video.generation.google.google_search import GoogleSearch
from yta_multimedia.video.generation.google.youtube_search import YoutubeSearch
from typing import Union


class PremadeElementBuilder(ElementBuilder):
    @classmethod
    def build_custom_from_premade_name(cls, premade_name: str, **parameters):
        ElementParameterValidator.validate_premade_name(premade_name)

        premade_class = None
        if premade_name == SegmentType.GOOGLE_SEARCH.value:
            premade_class = GoogleSearch
        elif premade_name == SegmentType.YOUTUBE_SEARCH.value:
            premade_class = YoutubeSearch
        else:
            raise Exception(f'The provided "premade_name" parameter {premade_name} is not a valid premade name.')

        return cls.build_custom(premade_class, **parameters)

    @classmethod
    def build_custom(cls, premade, **parameters):
        # TODO: Make the premades implement an abstract class named
        # 'Premade' to be able to detect them as subclasses
        return premade(**parameters).generate()

    @classmethod
    def build(cls, text: str, duration: Union[float, int]):
        """
        Basic example to test that the building process and
        the class are working correctly.

        TODO: Remove this in the future when 'custom' is 
        working perfectly.
        """
        ElementParameterValidator.validate_text(text)
        ElementParameterValidator.validate_duration(duration)

        return GoogleSearch(text).generate()


from youtube_autonomous.elements.builder.element_builder import ElementBuilder
from youtube_autonomous.segments.enums import SegmentField
from youtube_autonomous.elements.validator.element_parameter_validator import ElementParameterValidator


class AIVideoElementBuilder(ElementBuilder):
    """
    Builder to obtain videos generated with AI.
    """
    @classmethod
    def build_from_segment(cls, segment: dict):
        keywords = segment.get(SegmentField.KEYWORDS.value, None)
        # TODO: I should always have 'calculated_duration' when duration
        # has been processed
        duration = segment.get('calculated_duration', None)

        return cls.build(keywords, duration)

    @classmethod
    def build(cls, keywords: str, duration: Union[float, int]):
        ElementParameterValidator.validate_keywords(keywords)
        ElementParameterValidator.validate_duration(duration)

        raise Exception('Functionality not implemented yet.')

        # TODO: Build a video from some AI
        # video = create_ai_video(keywords, duration)

        return video


from youtube_autonomous.segments.enhancement.enhancement_element import EnhancementElement, EnhancementElementMode
from youtube_autonomous.segments.builder.youtube.youtube_downloader import YoutubeDownloader
from yta_multimedia.video.utils import parse_parameter_as_moviepy_clip
from yta_general_utils.downloader.audio import download_audio
from moviepy.editor import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, AudioFileClip
from typing import Union


class SoundEnhancementElement(EnhancementElement):
    @classmethod
    def get_valid_modes(cls):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return [EnhancementElementMode.OVERLAY]
    
    @classmethod
    def get_default_mode(cls):
        """
        The EnhancementElementMode enum defined as value by default.
        """
        return EnhancementElementMode.OVERLAY

    @property
    def valid_modes(self):
        """
        The EnhancementElementMode enums accepted for this type of
        EnhancementElement.

        See this link: https://www.notion.so/Segment-types-configuration-107f5a32d46280beb793d84a2af7f75e?pvs=4
        """
        return SoundEnhancementElement.get_valid_modes()

    def _build(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        youtube_downloader = YoutubeDownloader()

        filename = ''
        if self.filename:
            filename = self.filename
        elif self.url:
            filename = download_audio(self.url)
        elif self.keywords:
            youtube_downloader.activate_ignore_repeated()
            filename = youtube_downloader.download_sound_audio(self.keywords)
            youtube_downloader.deactivate_ignore_repeated()

        enhancement_clip = AudioFileClip(filename)

        super().recalculate_duration(enhancement_clip, video)

        return enhancement_clip

    def apply(self, video: Union[VideoFileClip, CompositeVideoClip, ImageClip, ColorClip, str]):
        video = parse_parameter_as_moviepy_clip(video)

        enhancement_clip = self._build(video)

        video = super().apply(enhancement_clip, video)

        return video